/*
    C-64 hires specific things
    
    Changelog:
    
    1.5.2106: Asm exporter
    30.4.2016: Saving/loading Art Studio (art/aas)
    30.4.2016: Saving/loading Doodle (dd/ddl)
    29.4.2016: Saving prg implemented, added populate() to create bitmap and colormap
    27.4.2016: Consider the aspect ratio
    26.4.2016: The first version
*/

class C64Hires extends Machine
{
    // Pepto
    final int rgb[]={#000000,#FFFFFF,#68372B,#70A4B2,
                     #6F3D86,#588D43,#352879,#B8C76F,
                     #6F4F25,#433900,#9A6759,#444444,
                     #6C6C6C,#9AD284,#6C5EB5,#959595};
                     
    Button save_prg_b,
           load_doodle_b,
           save_doodle_b,
           load_artstudio_b,
           save_artstudio_b,
           save_asm_b;
    
    C64Hires()
    {
        number=C64_HIRES;
        name="C-64 Hires";
        
        xsize=320;
        ysize=200;
        blox=bloy=gridx=gridy=8;
        
        renderx=xsize*prefs.zoom;
        rendery=ysize*prefs.zoom;
        if(prefs.zoom>1)
        {
            switch(prefs.aspect)
            {
                case PAL:  rendery=ysize*prefs.zoom*18/16; break;
                case NTSC: rendery=ysize*prefs.zoom*4/3; break;
                default:   rendery=ysize*prefs.zoom; break;
            }
        }
        
        colorselheight=prefs.UIWIDTH-2*prefs.BWIDTH;
        
        border=14;
        maxborder=15;
        pen=6;
        errorcolor=#10bb10;
        gridcolor=#205050;
        
        super.rgb=rgb;
    }
    
    // Fill colormap and bitmap arrays (two-color 8x8 blocks here)
    void populate()
    {
        int bx=xsize/blox,
            by=ysize/bloy;
        
        colormap=new int[bx*by];
        bitmap=new int[bx*by*8];
        
        pic.loadPixels();
        for(int y=0,i=0;y<ysize;y+=bloy)
            for(int x=0;x<xsize;x+=blox,i++)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
                
                colormap[i]=count[0][0]*16; // fg = most common color
                if(count.length>1)
                    colormap[i]+=count[1][0];
                    
                for(int j=0;j<8;j++) // Create the bitmap
                {
                    int bits=0;
                    for(int k=0;k<8;k++)
                        if(pix[j*8+k]==count[0][1])
                            bits+=(0x80>>k);
                    
                    bitmap[i*8+j]=bits;
                }
            }
    }
    
    // Generic hires decode
    void decode(byte b[],int coff,int boff,int target[])
    {
        for(int y=0;y<ysize;y+=bloy)
        {
            for(int x=0;x<xsize;x+=blox,coff++)
            {
                int fg=rgb[(((int)b[coff])&0xff)>>4],
                    bg=rgb[(int)b[coff]&0xf];
                
                for(int i=0;i<bloy;i++,boff++) // Char lines
                {
                    for(int j=0;j<blox;j++)
                    {
                        if((b[boff]&(0x80>>j))!=0)
                            target[x+j+(y+i)*xsize]=fg;
                        else
                            target[x+j+(y+i)*xsize]=bg;
                    }
                }
            }
        }
    }
    
    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_doodle_b=new Button(x,super.addloadbuttons(x,y),"load doodle");
        load_artstudio_b=new Button(x,load_doodle_b.next,"load art/aas");
        return load_artstudio_b.next;
    }
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_prg_b=new Button(x,super.addsavebuttons(x,y),"save prg");
        save_doodle_b=new Button(x,save_prg_b.next,"save doodle");
        save_artstudio_b=new Button(x,save_doodle_b.next,"save art/aas");
        save_asm_b=new Button(x,save_artstudio_b.next,"save asm");
        return save_asm_b.next;
    }
   
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_prg_b.mouseover() || save_doodle_b.mouseover() || save_artstudio_b.mouseover() || save_asm_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_prg_b.mouseover()) // PRG
                    {
                        String s=fileselector(prefs.path,"PRG files",new String[] {"prg"},SAVE);
                        if(s!=null)
                            save_prg(s);
                    }
                    if(save_doodle_b.mouseover()) // Doodle
                    {
                        String s=fileselector(prefs.path,"Doodle files",new String[] {"dd","ddl"},SAVE);
                        if(s!=null)
                            save_doodle(s);
                    }
                    if(save_artstudio_b.mouseover()) // Art Studio
                    {
                        String s=fileselector(prefs.path,"Art Studio files",new String[] {"aas","art"},SAVE);
                        if(s!=null)
                            save_artstudio(s);
                    }
                    if(save_asm_b.mouseover()) // Assembler
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(load_doodle_b.mouseover())
        {
            String s=fileselector(prefs.path,"Doodle files",new String[] {"dd","ddl"},LOAD);
            if(s!=null)
                load_doodle(s);
        }
        if(load_artstudio_b.mouseover())
        {
            String s=fileselector(prefs.path,"Art Studio files",new String[] {"aas","art"},LOAD);
            if(s!=null)
                load_artstudio(s);
        }
    }
    
    // Save PRG
    void save_prg(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/c64-hires.prg");
        
        populate(); // Create the bitmap/colormap  
        
        // Replace some bytes
        b[0x2167]=(byte)border;
        for(int i=0;i<bitmap.length;i++)
            b[0x227+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[0x2169+i]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save Doodle
    void save_doodle(String name)
    {
        if(pic==null)
            return;

        byte b[]=new byte[9218];
        
        populate(); // Create the bitmap/colormap
        
        // Replace some bytes
        b[0]=0;
        b[1]=0x5c;
        for(int i=0;i<colormap.length;i++)
            b[2+i]=(byte)colormap[i];
        for(int i=0;i<bitmap.length;i++)
            b[1026+i]=(byte)bitmap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Save Doodle
    void load_doodle(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<9026)
        {
            message("Can't open file: "+name);
            return;
        }
        message("Loaded: "+name);
        
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        decode(b,2,1026,sourcepic.pixels);
           
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }
    
    // Save Art Studio
    void save_artstudio(String name)
    {
        if(pic==null)
            return;

        byte b[]=new byte[9002];
        
        populate(); // Create the bitmap/colormap
        
        // Replace some bytes
        b[0]=0;
        b[1]=0x20;
        for(int i=0;i<bitmap.length;i++)
            b[2+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[8002+i]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Save Art Studio
    void load_artstudio(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<9002)
        {
            message("Can't open file: "+name);
            return;
        }
        message("Loaded: "+name);
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        decode(b,8002,2,sourcepic.pixels);
           
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }
    
    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
        
        PrintWriter f=createWriter(name);
        
        f.print("bitmap");
        printarr(f,bitmap,".byte","$");
        f.println();
        f.print("scrram");
        printarr(f,colormap,".byte","$");
        f.println();
        f.println("border\t.byte\t$"+hex(border,2));

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
