/*
    C-64 multicolor specific things
    
    Changelog:

    18.5.2016: Smarter algorithm for bg color finding
    3.5.2016: RMB works for bg color selection too
    3.5.2016: Shift-click for selecting the bg color and a force button
    1.5.2016: Asm exporter
    1.5.2016: Show bg in the color selector
    1.5.2016: Keep the bg found on images instead of calculating a new one (forcedbg)
    1.5.2016: Support for Advanced Art Studio (OCP) loading/saving
    30.4.2016: Possibly supports compressed .gg Koala files now
    30.4.2016: Koala loading and saving, accept both .koa and .kla
    30.4.2016: Multicolor prg output
    29.4.2016: Added multicolor-specific check()
    28.4.2016: The first version
*/

class C64Multicolor extends Machine
{
    // Pepto
    final int rgb[]={#000000,#FFFFFF,#68372B,#70A4B2,
                     #6F3D86,#588D43,#352879,#B8C76F,
                     #6F4F25,#433900,#9A6759,#444444,
                     #6C6C6C,#9AD284,#6C5EB5,#959595};
                     
    int bg,
        screenram[];
        
    boolean forcedbg;
                     
    Button save_prg_b,
           load_koala_b,
           save_koala_b,
           load_ocp_b,
           save_ocp_b,
           save_asm_b;
           
    Checkbutton fix_bg_b;
    
    C64Multicolor()
    {
        number=C64_MULTICOLOR;
        name="C-64 Multicolor";
        
        xsize=160;
        ysize=200;
        blox=gridx=4;
        bloy=gridy=8;
        
        renderx=xsize*prefs.zoom*2; // Pixels are about 2x as wide as high
        rendery=ysize*prefs.zoom;
        if(prefs.zoom>1)
        {
            switch(prefs.aspect)
            {
                case PAL:  rendery=ysize*prefs.zoom*18/16; break;
                case NTSC: rendery=ysize*prefs.zoom*4/3; break;
                default:   rendery=ysize*prefs.zoom; break;
            }
        }
        
        colorselheight=prefs.UIWIDTH-2*prefs.BWIDTH;
        
        border=14;
        maxborder=15;
        pen=6;
        bg=0;
        forcedbg=false;
        
        errorcolor=#10bb10;
        gridcolor=#205050;
        
        super.rgb=rgb;
    }

    // Draw extras on the colorsel
    void drawcolorsel(int px,int py)
    {
        super.drawcolorsel(px,py);
        
        int wide=(prefs.UIWIDTH-2*prefs.BWIDTH)/4;
        
        // Loop needed some day if we allow forcing the bg
        for(int y=0,i=0;y<rgb.length/4;y++)
            for(int x=0;x<4;x++,i++)
            {
                if(mousePressed && (shift || mouseButton==RIGHT)) // Set bg
                {
                    if(inside(px+x*wide,py+y*wide, px+x*wide+wide,py+y*wide+wide))
                        bg=i;
                }
                if(i==bg)
                {
                    fill(rgb[i]^0x808080);
                    rect(px+x*wide+8,py+y*wide+8, wide-16,wide-16);
                }
            }
        
        noStroke();
        fill(prefs.UICOLOR);
        rect(px,py+colorselheight+2,prefs.UIWIDTH-prefs.BWIDTH,prefs.UIROW);
        fill(prefs.UITEXT);
        text("bg: "+str(bg)+" bdr: "+str(border),px,py+wide*rgb.length/4+prefs.UIROW-7);
    }

    // Accept both 320x200 and 160x200
    boolean correctsize(int x,int y)
    {
        if((x==xsize || x==xsize*2) && y==ysize)
            return true;
        else
            return false;
    }
    
    // Check multicolor's 4-color limits
    boolean check(boolean fix)
    {
        boolean errorfound=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        pic.loadPixels();
        
        // Find the most common 4-block color, which will become the bg
        if(!forcedbg)
        {
            // The simple but nonoptimal way
            //bg=countcolors(pic.pixels,rgb)[0][0];
            
            int freqs[]=new int[rgb.length];
            
            for(int y=0;y<ysize;y+=bloy)
                for(int x=0;x<xsize;x+=blox)
                {
                    int pix[]=getpixels(x,y,blox,bloy);
                    int count[][]=countcolors(pix,rgb);
                    
                    if(count.length>3) // Four color blocks
                    {
                        for(int i=0;i<count.length;i++) // Add frequency for each color
                            freqs[count[i][0]]++;
                    }                    
                }
                
            // Now we have freqs for each color in 3+ color blocks. Let's find the most common.
            int largest=0;
            for(int i=0;i<freqs.length;i++)
                if(freqs[i]>largest)
                {
                    largest=freqs[i];
                    bg=i;
                }
        }
        int brgb=rgb[bg];

        for(int y=0;y<ysize;y+=bloy) // Walk through every block
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
 
                if(count.length>3) // More than three?
                {
                    boolean flagga=true;
                    if(count.length==4) // Maybe one of them is the bg?
                    {
                        for(int i=0;i<count.length;i++)
                            if(brgb==count[i][1]) // Yes it is
                                flagga=false;
                    }
                    
                    if(flagga) // More than three and one is not the bg
                    {
                        errorfound=true;
                        if(fix)
                        {
                            // Use the three most common colors + bg for the remaining ones
                            boolean oneisbg=false;
                            for(int i=0;i<4;i++)
                                if(count[i][1]==brgb)
                                    oneisbg=true;
                                    
                            if(!oneisbg)
                                count[3][1]=brgb;
                            
                            for(int i=0;i<pix.length;i++)
                            {
                                int dist=1000000,best=0;
                                for(int j=0;j<4;j++)
                                {
                                    if(rgbdist(pix[i],count[j][1])<dist)
                                    {
                                        dist=rgbdist(pix[i],count[j][1]);
                                        best=count[j][1];
                                    }
                                }
                                pix[i]=best;
                            }
                            setpixels(x,y,blox,bloy,pix);
                        }
                        else
                            setpixels(x,y,blox,bloy,errorcolor);
                    }
                }
            }
        }
        pic.updatePixels();
        
        if(errorfound)
        {
            message("Color clash(es) found");
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }
    
    // Compile the bit- and colormaps
    void populate()
    {
        int off=0;
        
        bitmap=new int[40*25*8];
        colormap=new int[40*25];
        screenram=new int[40*25];
        
        pic.loadPixels();
        
        // Find the most common color, which will become the fixed bg (need to rethink later)
        if(!forcedbg)
            bg=countcolors(pic.pixels,rgb)[0][0];
        int brgb=rgb[bg];

        for(int y=0;y<ysize;y+=bloy) // Walk through every block
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
                int bloi=x/blox+40*(y/bloy);
                
                if(count.length>4)
                {
                    message("Can't save with color clashes.");
                    return;
                }

                // Form the color maps
                colormap[bloi]=screenram[bloi]=0;
                int sr1=ILLEGAL,sr2=ILLEGAL,cm=ILLEGAL;
                
                for(int i=0,j=0;i<count.length;i++)
                {
                    if(count[i][1]!=brgb)
                    {
                        if(j==0) {screenram[bloi]=count[i][0]*16; sr1=count[i][1];}
                        if(j==1) {screenram[bloi]+=count[i][0];   sr2=count[i][1];}
                        if(j==2) {colormap[bloi]=cm=count[i][0];  cm=count[i][1];}
                        j++;
                    }
                }
                
                // Calculate the bitmap
                for(int i=0;i<bloy;i++,off++)
                {
                    bitmap[off]=0;
                    for(int j=0;j<4;j++) // Check which of the 4 choices fits best
                    {
                        int bits=0,
                            diff,
                            kolor=pix[j+i*4];
                        
                        diff=rgbdist(kolor,brgb);
                        if(rgbdist(kolor,sr1)<diff)
                        {
                            diff=rgbdist(kolor,sr1);
                            bits=1;
                        }
                        if(rgbdist(kolor,sr2)<diff)
                        {
                            diff=rgbdist(kolor,sr2);
                            bits=2;
                        }
                        if(rgbdist(kolor,cm)<diff)
                            bits=3;
                            
                        bitmap[off]+=((bits<<6)>>j*2);
                    }
                }
            }
        }
    }
    
    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_koala_b=new Button(x,super.addloadbuttons(x,y),"load koala");
        load_ocp_b=new Button(x,load_koala_b.next,"load ocp");
        return load_ocp_b.next;
    }
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_prg_b=new Button(x,super.addsavebuttons(x,y),"save prg");
        save_koala_b=new Button(x,save_prg_b.next,"save koala");
        save_ocp_b=new Button(x,save_koala_b.next,"save ocp");
        save_asm_b=new Button(x,save_ocp_b.next,"save asm");
        return save_asm_b.next;
    }
    // Even an extra button
    int addextrabuttons(int x,int y)
    {
        fix_bg_b=new Checkbutton(x,super.addextrabuttons(x,y)+prefs.UIGAP,"fixed bg");
        return fix_bg_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_prg_b.mouseover() || save_koala_b.mouseover() || save_ocp_b.mouseover() || save_asm_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                boolean forced=forcedbg; // Force it temporarily to keep the calculated fg/bg
                forcedbg=true;
                
                if(check(false)) // Seems to be fine
                {
                    if(save_prg_b.mouseover()) // PRG
                    {
                        String s=fileselector(prefs.path,"PRG files",new String[] {"prg"},SAVE);
                        if(s!=null)
                            save_prg(s);
                    }
                    if(save_koala_b.mouseover()) // Koala
                    {
                        String s=fileselector(prefs.path,"Koala files",new String[] {"koa","kla"},SAVE);
                        if(s!=null)
                            save_koala(s);
                    }
                    if(save_ocp_b.mouseover()) // Acvanced Art Studio ocp
                    {
                        String s=fileselector(prefs.path,"Advanced Art Studio files",new String[] {"ocp"},SAVE);
                        if(s!=null)
                            save_ocp(s);
                    }
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
                forcedbg=forced;
            }
        }
        
        if(load_koala_b.mouseover())
        {
            String s=fileselector(prefs.path,"Koala files",new String[] {"koa","kla","gg"},LOAD);
            if(s!=null)
                load_koala(s);
        }
        if(load_ocp_b.mouseover())
        {
            String s=fileselector(prefs.path,"Advanced Art Studio files",new String[] {"ocp"},LOAD);
            if(s!=null)
                load_ocp(s);
        }
        
        if(fix_bg_b.mouseover())
        {
            forcedbg=!forcedbg;
            fix_bg_b.checked=forcedbg;
        }
    }
    
    // Save PRG
    void save_prg(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/c64-multicolor.prg");
        
        populate(); // Create bitmap/colormap/screen ram
        
        // Replace some bytes
        b[0x2179]=(byte)border;
        b[0x217a]=(byte)bg;
        
        for(int i=0;i<bitmap.length;i++)
            b[0x239+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[0x2563+i]=(byte)colormap[i];
        for(int i=0;i<screenram.length;i++)
            b[0x217b+i]=(byte)screenram[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save Koala
    void save_koala(String name)
    {
        if(pic==null)
            return;

        byte b[]=new byte[10003];
        
        populate(); // Create bitmap/colormap/screen ram  
        
        b[0]=0;
        b[1]=0x60;
        
        int off=2;
        for(int i=0;i<bitmap.length;i++)
            b[off++]=(byte)bitmap[i];
        for(int i=0;i<screenram.length;i++)
            b[off++]=(byte)screenram[i];
        for(int i=0;i<colormap.length;i++)
            b[off++]=(byte)colormap[i];
        
        b[off]=(byte)bestmatch(bg,rgb);
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Load Koala
    void load_koala(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null)
        {
            message("Can't open file: "+name);
            return;
        }
        if(!getext(name).equalsIgnoreCase(".gg") && b.length<10003)
        {
            message("Malformed file: "+name);
            return;
        }
        message("Loaded: "+name);
        
        if(getext(name).equalsIgnoreCase(".gg")) // Completely untested
        {
            byte tmp[]=b;
            b=new byte[10003+1000]; // Some extra space just in case...
            
            int off=0,doff=0;
            while(true)
            {
                int c=((int)tmp[off])&0xff;
                if(c==0xfe) // compressed
                {
                    off++;
                    byte byy=tmp[off++];
                    int count=((int)tmp[off++])&0xff;
                    
                    for(int i=0;i<count;i++)
                        b[doff++]=byy;
                }
                else
                    b[doff++]=tmp[off++];
                    
                if(off>=tmp.length)
                    break;
            }
        }
        
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        bg=b[10002]&0xf;
        
        for(int y=0,off=2,boff=0;y<ysize;y+=bloy) // Blocks
            for(int x=0;x<xsize;x+=blox,boff++)
            {
                for(int i=0;i<bloy;i++,off++) // Block rows
                {
                    for(int j=0;j<blox;j++)
                    {
                        int bits=(((int)b[off])>>(6-j*2))&3, // Bitmap decides where the color comes from
                            kolor=bg;
                            
                        if(bits==1)
                            kolor=(b[8002+boff]>>4)&0xf; // Screen RAM hi nibble
                        if(bits==2)
                            kolor=b[8002+boff]&0xf;      // Screen RAM lo nibble
                        if(bits==3)
                            kolor=b[9002+boff]&0xf;      // Color RAM
                        
                        sourcepic.pixels[x+j+(y+i)*xsize]=rgb[kolor];
                    }
                }
            }
        
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }
    
    // Save OCP
    void save_ocp(String name)
    {
        if(pic==null)
            return;

        byte b[]=new byte[10018];
        
        populate(); // Create bitmap/colormap/screen ram
        
        b[0]=0;
        b[1]=0x20;
        
        int off=2;
        for(int i=0;i<bitmap.length;i++)
            b[off++]=(byte)bitmap[i];
        for(int i=0;i<screenram.length;i++)
            b[off++]=(byte)screenram[i];

        b[off++]=(byte)border;
        b[off++]=(byte)bg;
        off+=14;
        
        for(int i=0;i<colormap.length;i++)
            b[off++]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Load OCP
    void load_ocp(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<10018)
        {
            message("Can't open file: "+name);
            return;
        }

        message("Loaded: "+name);
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        // Need to consolidate this sometime with the same loop from koala_save()
        border=b[9002]&0xf;
        bg=b[9003]&0xf;
        
        for(int y=0,off=2,boff=0;y<ysize;y+=bloy) // Blocks
            for(int x=0;x<xsize;x+=blox,boff++)
            {
                for(int i=0;i<bloy;i++,off++) // Block rows
                {
                    for(int j=0;j<blox;j++)
                    {
                        int bits=(((int)b[off])>>(6-j*2))&3, // Bitmap decides where the color comes from
                            kolor=bg;
                            
                        if(bits==1)
                            kolor=(b[8002+boff]>>4)&0xf; // Screen RAM hi nibble
                        if(bits==2)
                            kolor=b[8002+boff]&0xf;      // Screen RAM lo nibble
                        if(bits==3)
                            kolor=b[9018+boff]&0xf;      // Color RAM
                        
                        sourcepic.pixels[x+j+(y+i)*xsize]=rgb[kolor];
                    }
                }
            }
        
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }
    
    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap/screen ram
        
        PrintWriter f=createWriter(name);
        
        f.print("bitmap");
        printarr(f,bitmap,".byte","$");
        f.println();
        f.print("scrram");
        printarr(f,screenram,".byte","$");
        f.println();
        f.print("cmap");
        printarr(f,colormap,".byte","$");
        f.println();
        f.println("bg\t.byte\t$"+hex(bg,2));
        f.println("border\t.byte\t$"+hex(border,2));

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
