/*
    Panasonic JR-200 specific things
    
    Changelog:
    
    20.5.2016: CJR export done
    20.5.2016: populate() and asm export
    20.5.2016: Around 6x faster force(), precalculate bit counting and give error message about clashes and blocks
    20.5.2016: A lot better but too slow forcing in place
    19.5.2016: Block checking and simple forcing implemented
    18.5.2016: Support for loading Spectrum SCR files
    18.5.2016: Color checking using the superclass, still need block count
    18.5.2016: Startup colors green and cyan now
    18.5.2016: First created
*/

class PanasonicJR200 extends Machine
{
    // Simple 8-bit RGB
    final int rgb[]={#000000,#0000FF,#FF0000,#FF00FF,
                     #00FF00,#00FFFF,#FFFF00,#FFFFFF};
                     
    int indexmap[],
        xorcount[];
                     
    Button load_scr_b,
           save_cjr_b,
           save_asm_b;
    
    PanasonicJR200()
    {
        number=PANASONIC_JR200;
        name="Panasonic JR-200";
        
        xsize=256;
        ysize=192;
        blox=bloy=gridx=gridy=8;

        // Let's just assume they are square pixels        
        renderx=xsize*prefs.zoom;
        rendery=ysize*prefs.zoom;

        colorselheight=(prefs.UIWIDTH-2*prefs.BWIDTH)/2;
        
        border=4;
        maxborder=7;
        pen=5;
        
        super.rgb=rgb;
        
        // Precalc for samebits()
        xorcount=new int[256];
        for(int i=0;i<256;i++)
        {
            int n=0;
            for(int j=0;j<8;j++)
                if((i&(0x80>>j))!=0)
                    n++;
                
            xorcount[i]=8-n;
        }
    }
       
    
    // Color clash + different block counter (must be <= 320)
    boolean check(boolean fix)
    {
        boolean errorfound,toomany=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        // The superclass checks clashes already
        errorfound=!super.check(fix);
        if(errorfound && !fix)
            return false;

        // The colors are ok by now one way or another, check block count next
        pic.loadPixels();
        ArrayList<int[]> blocks=new ArrayList();
        int idx[]=new int[(ysize/bloy)*(xsize/blox)];
        
        // Go through each and add it to "blocks" if it's new
        for(int y=0,i=0;y<ysize;y+=bloy)
        {
            for(int x=0;x<xsize;x+=blox,i++)
            {
                int pix[]=getpixels(x,y, blox,bloy),
                    bits[]=bits(pix);
                
                boolean found=false;
                
                for(int j=0;j<blocks.size();j++)
                {
                    int blk[]=blocks.get(j),
                        result=samebits(bits,blk);

                    if(result==0 || result==64) // Same bitmap found already before
                    {
                        found=true;
                        idx[i]=j;
                    }
                }
                
                if(!found) // A new block, let's see
                {
                    if(blocks.size()>=320) // Maximum reached
                    {
                        toomany=true;
                        if(!fix)
                            setpixels(x,y, blox,bloy, errorcolor);
                            
                        blocks.add(bits);
                        idx[i]=blocks.size()-1;
                    }
                    else // Still room
                    {
                        blocks.add(bits);
                        idx[i]=blocks.size()-1;
                    }
                }
            }
        }
        
        int blockcount=blocks.size();
        
        // Too many: need to find the best blocks for the pic
        if(fix && blocks.size()>320)
        {
            // Reduce to 320
            int prev=-1;
            while(blocks.size()>320)
            {
                int pair1=0,pair2=0,
                    diff=-1;
                
                // Find the nearest bitmaps
                for(int i=0;i<blocks.size();i++)
                {
                    for(int j=i+1;j<blocks.size();j++)
                    {
                        int result=samebits(blocks.get(i),blocks.get(j));
                        if(result<32)
                            result=64-result;
                            
                        if(result>diff)
                        {
                            diff=result;
                            pair1=i;
                            pair2=j;
                        }
                    }
                    if(diff==prev) // Can't find better than this
                        break;
                }
                prev=diff;
                
                // Find which one is more common
                int fr1=0,fr2=0;
                for(int i=0;i<idx.length;i++)
                {
                    if(idx[i]==pair1)
                        fr1++;
                    if(idx[i]==pair2)
                        fr2++;
                }
                if(fr2>fr1) // And swap if needed
                {
                    blocks.set(pair1,blocks.get(pair2));
                }
                
                // The other needs to go
                blocks.remove(pair2);
                for(int i=0;i<idx.length;i++) // Change all affected charnums too
                {
                    if(idx[i]==pair2) // Reassign
                    {
                        idx[i]=pair1;
                    }
                    else // Move down
                    {
                        if(idx[i]>pair2)
                            idx[i]--;
                    }
                }
            }

            // Fix blocks to their new bitmaps
            for(int y=0,i=0;y<ysize;y+=bloy)
            {
                for(int x=0;x<xsize;x+=blox,i++)
                {   
                    int pix[]=getpixels(x,y, blox,bloy),
                        bits[]=bits(pix),
                        count[][]=countcolors(pix,rgb),
                        tbits[]=blocks.get(idx[i]);
                    
                    int col1=count[0][1],
                        col2=0; 
                    if(count.length>1)
                        col2=count[1][1];
                        
                    boolean invert=samebits(bits,tbits)<32;
                    
                    for(int j=0;j<bloy;j++)
                    {
                        for(int k=0;k<blox;k++)
                        {
                            if((tbits[j]&(0x80>>k))==0)
                            {
                                if(invert)
                                    pix[j*blox+k]=col2;
                                else
                                    pix[j*blox+k]=col1;
                            }
                            else
                            {
                                if(invert)
                                    pix[j*blox+k]=col1;
                                else
                                    pix[j*blox+k]=col2;
                            }
                        }
                    }
                        
                    setpixels(x,y, blox,bloy, pix);
                }
            }
        }
        
        pic.updatePixels();
        
        if(errorfound)
        {
            if(toomany)
                message("Color clash(es) and too many blocks: "+str(blockcount));
            else
                message("Color clash(es) found");
            return false;
        }
        else
        {
            if(toomany)
            {
                message("Too many different blocks: "+str(blockcount));
                return false;
            }
            else
            {
                message("Colors ok");
                return true;
            }
        }
    }
    
    // Find number of same bits
    int samebits(int a[],int b[])
    {
        int same=0;
        for(int i=0;i<a.length;i++)
        {
            same+=xorcount[a[i]^b[i]];

// Old non-lookup way
//            for(int j=0;j<8;j++)
//                if((a[i]&(0x80>>j)) == (b[i]&(0x80>>j)))
//                    same++;
        }
        
        return same;
    }
    // Return the bits: most common color bg (0), others fg (1)
    int[] bits(int a[])
    {
        int count[][]=countcolors(a,rgb),
            bits[]=new int[8];

        // Count all            
        for(int i=0;i<8;i++)
            for(int j=0;j<8;j++)
                if(a[i*8+j]!=count[0][1])
                    bits[i]+=(0x80>>j);
        
        return bits;
    }
    
    // Create chars / screen in JR200 format. Needs to be checked before.
    void populate()
    {
        int bx=xsize/blox,
            by=ysize/bloy;
        
        colormap=new int[bx*by];
        bitmap=new int[320*8];
        indexmap=new int[bx*by];

        // The colors and blocks should be fine by now
        pic.loadPixels();
        ArrayList<int[]> blocks=new ArrayList();
        int idx[]=new int[(ysize/bloy)*(xsize/blox)];
        
        // Go through each and add it to "blocks" if it's new
        for(int y=0,i=0;y<ysize;y+=bloy)
        {
            for(int x=0;x<xsize;x+=blox,i++)
            {
                int pix[]=getpixels(x,y, blox,bloy),
                    bits[]=bits(pix);
                
                boolean found=false,
                        invert=false;
                
                for(int j=0;j<blocks.size();j++)
                {
                    int blk[]=blocks.get(j),
                        result=samebits(bits,blk);

                    if(result==0 || result==64) // Same bitmap found already before
                    {
                        found=true;
                        indexmap[i]=j;
                        if(result==0)
                            invert=true;
                        else
                            invert=false;
                    }
                }
                
                if(!found) // A new block, let's see
                {
                    if(blocks.size()==320) // Maximum reached - shouldn't happen tho
                    {
                        println("Something wrong. Too many blocks");
                    }
                    else // Still room
                    {
                        blocks.add(bits);
                        indexmap[i]=blocks.size()-1;
                    }
                }
                
                // Colors too
                int count[][]=countcolors(pix,rgb);
                if(invert)
                    colormap[i]=count[0][0];
                else
                    colormap[i]=count[0][0]<<3;
                if(count.length>1)
                {
                    if(invert)
                        colormap[i]+=count[1][0]<<3;
                    else
                        colormap[i]+=count[1][0];
                }
                
                // Need bit 6 to go beyond 256 chars
                if(indexmap[i]>255)
                {
                    indexmap[i]-=256;
                    if(indexmap[i]>31)
                        indexmap[i]+=96; // Deal with the gap
                    
                    colormap[i]+=0x40;
                }
            }
        }
        
        // And then the bitmap
        for(int i=0;i<blocks.size();i++)
            for(int j=0;j<bloy;j++)
                bitmap[i*8+j]=blocks.get(i)[j];
    }

    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
       load_scr_b=new Button(x,super.addloadbuttons(x,y),"load scr");
       return load_scr_b.next;
    }
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_cjr_b=new Button(x,super.addsavebuttons(x,y),"save cjr");
        save_asm_b=new Button(x,save_cjr_b.next,"save asm");
        return save_asm_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_asm_b.mouseover() || save_cjr_b.mouseover())
        {
            if(pic==null)
                message("Can't save an empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                    if(save_cjr_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"CJR files",new String[] {"cjr"},SAVE);
                        if(s!=null)
                            save_cjr(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(load_scr_b.mouseover())
        {
            String s=fileselector(prefs.path,"Spectrum SCR files",new String[] {"scr"},LOAD);
            if(s!=null)
                load_scr(s);
        }
    }

    // Load Spectrum SCR
    void load_scr(String name)
    {
        // No need to duplicate code, as Spectrum can load those images already
        ZXSpectrum speccy=new ZXSpectrum();
        
        speccy.load_scr(name);
        sourcepic=speccy.sourcepic;
        pic=speccy.pic;
        colorcorrect();
    }

    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
       
        PrintWriter f=createWriter(name);
        
        f.println("border\tdb\t$"+hex(border,2));
        f.println();
        f.print("index");
        printarr(f,indexmap,"db","$");
        f.println();
        f.print("color");
        printarr(f,colormap,"db","$");
        f.println();
        f.print("bitmap");
        printarr(f,bitmap,"db","$");
        f.println();

        f.flush();
        f.close();
        message("Saved: "+name);
    }
    
    // Save CJR emulator tape
    void save_cjr(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/jr-200.bin");
        
        populate(); // Create bitmap/colormap/screen ram
        
        // Replace some bytes
        b[117]=(byte)border;
        
        for(int i=0;i<indexmap.length;i++)
            b[118+i]=(byte)indexmap[i];
        for(int i=0;i<colormap.length;i++)
            b[886+i]=(byte)colormap[i];
        for(int i=0;i<bitmap.length;i++)
            b[1654+i]=(byte)bitmap[i];
        
        saveBytes(name,makecjr(b,"image",0x1000));
        message("Saved: "+name);
    }
    // Turn binary into CJR with blocks, checksums and all
    byte[] makecjr(byte data[],String name,int loadaddr)
    {
        int off=0,
            dataoff=0,
            blknum=0,
            length=data.length;
            
        byte dummy[]=new byte[100000], // Oh well, big enough
             blk[]=new byte[256];
            
        // 1st blk is different
        for(int i=0;i<16 && i<name.length();i++)
            blk[i]=(byte)name.charAt(i);
        blk[16]=1;
        blk[17]=0;
        for(int i=0;i<8;i++)
            blk[18+i]=(byte)255;
        off=createblk(dummy,off, blk,26, 0xffff,blknum);
        blknum++;
        
        // Data blocks next
        while(length!=0)
        {
            int blen=256;
            if(length<256)
                blen=length;
            
            for(int i=0;i<blen;i++)
                blk[i]=data[dataoff++];
            off=createblk(dummy,off, blk,blen, loadaddr,blknum);
            
            length-=blen;
            loadaddr+=blen;
            blknum++;
        }
        
        // Last is different too
        dummy[off+0]=2;
        dummy[off+1]=42;
        dummy[off+2]=(byte)255;
        dummy[off+3]=(byte)255;
        dummy[off+4]=(byte)(loadaddr>>8);
        dummy[off+5]=(byte)(loadaddr&0xff);
        
        off+=6;
        
        // Resize and return
        byte result[]=new byte[off];
        System.arraycopy(dummy,0,result,0,off);
        return result;
    }
    // Create one block with checksum etc.
    int createblk(byte dest[],int off,byte bytes[],int len,int addr,int blknum)
    {
        int sum=0;

        // Block header
        dest[off+0]=2;
        dest[off+1]=42;
        dest[off+2]=(byte)blknum;
        dest[off+3]=(byte)(len&0xff);
        dest[off+4]=(byte)(addr>>8);
        dest[off+5]=(byte)(addr&0xff);
        
        // Data
        for(int i=0;i<len;i++)
            dest[off+6+i]=bytes[i];
            
        // Checksum
        for(int i=0;i<len+6;i++)
            sum+=((int)dest[off+i])&0xff;
        dest[off+len+6]=(byte)sum;

        return off+7+len;
    }
}
