/*
    MSX and ColecoVision Screen 2 specific things
    
    Changelog:
    
    24.5.2016: ColecoVision added as a curiosity (COL)
    24.5.2016: Take name table into account when loading SC2 images
    1.5.2016: Asm export (approximate Gen80 format)
    29.4.2016: Implemented COM and SC2 export
    27.4.2016: Implemented a preliminary color checker... and moved it to Machine
    27.4.2016: Consider the aspect ratio
    26.4.2016: The first version
*/

class MsxScreen2 extends Machine
{
    // Supposedly from the TI datasheet
    final int rgb[]={ILLEGAL,#000000,#21c842,#5edc78,
                     #5455ed,#7d76fc,#d4524d,#42ebf5,
                     #fc5554,#ff7978,#d4c154,#e6ce80,
                     #21b03b,#c95bba,#cccccc,#ffffff};
                     
    Button save_com_b,
           load_sc2_b,
           save_sc2_b,
           save_col_b,
           save_asm_b;
    
    MsxScreen2()
    {
        number=MSX_SCREEN2;
        name="MSX/ColecoVision Screen 2";

        xsize=256;
        ysize=192;
        blox=gridx=gridy=8;
        bloy=1;
        
        renderx=xsize*prefs.zoom;
        rendery=ysize*prefs.zoom;
        if(prefs.zoom>1)
        {
            if(prefs.aspect==PAL)
                renderx=xsize*prefs.zoom*12/10; // Approximate stretch (it IS flat!)
        }
       
        colorselheight=prefs.UIWIDTH-2*prefs.BWIDTH;
       
        border=4;
        maxborder=15;
        pen=4;
        errorcolor=#ffff33;
        gridcolor=#778822;

        super.rgb=rgb;
    }
    
    // Create a bitmap and the color maps
    void populate()
    {
        int boff=0,coff=0;
        
        colormap=new int[32*24*8];
        bitmap=new int[32*24*8];
        
        pic.loadPixels();
        for(int y=0;y<ysize;y+=8) // Walk through chars
            for(int x=0;x<xsize;x+=8)
            {
                for(int i=0;i<8;i++) // and rows
                {
                    int pix[]=getpixels(x,y+i,blox,bloy);
                    int count[][]=countcolors(pix,rgb);
                    
                    colormap[coff]=count[0][0]*16; // fg = most common color
                    if(count.length>1)
                        colormap[coff]+=count[1][0];
                    coff++;
                
                    int bits=0;
                    for(int k=0;k<8;k++)
                        if(pix[k]==count[0][1])
                            bits+=(0x80>>k);
                    
                    bitmap[boff++]=bits;
                }
            }
    }

    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_sc2_b=new Button(x,super.addloadbuttons(x,y),"load sc2");
        return load_sc2_b.next;
    }    
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_com_b=new Button(x,super.addsavebuttons(x,y),"save com");
        save_sc2_b=new Button(x,save_com_b.next,"save sc2");
        save_col_b=new Button(x,save_sc2_b.next,"save col");
        save_asm_b=new Button(x,save_col_b.next,"save asm");
        return save_asm_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_com_b.mouseover() || save_sc2_b.mouseover() || save_asm_b.mouseover() || save_col_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_com_b.mouseover()) // COM
                    {
                        String s=fileselector(prefs.path,"COM files",new String[] {"com"},SAVE);
                        if(s!=null)
                            save_com(s);
                    }
                    if(save_sc2_b.mouseover()) // SC2
                    {
                        String s=fileselector(prefs.path,"SC2 files",new String[] {"sc2"},SAVE);
                        if(s!=null)
                            save_sc2(s);
                    }
                    if(save_col_b.mouseover()) // ColecoVision COL
                    {
                        String s=fileselector(prefs.path,"COL files",new String[] {"col"},SAVE);
                        if(s!=null)
                            save_col(s);
                    }
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"s","asm"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(load_sc2_b.mouseover())
        {
            String s=fileselector(prefs.path,"SC2 files",new String[] {"sc2"},LOAD);
            if(s!=null)
                load_sc2(s);
        }
    }
    
    // Save COM
    void save_com(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/msx-screen2.com");
        
        populate(); // Create the bitmap/colormap  
        
        // Replace some bytes
        b[0x18f9]=(byte)border;
        for(int i=0;i<bitmap.length;i++)
            b[0xf9+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[0x18fa+i]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save SC2
    void save_sc2(String name)
    {
        int header[]={0xfe,0,0,0xff,0x37,0,0},
            off=0;
        
        if(pic==null)
            return;

        byte b[]=new byte[7+32*192*2+2048];
        
        populate(); // Create the bitmap/colormap
        
        for(int i=0;i<header.length;i++) // Header
            b[off++]=(byte)header[i];
        for(int i=0;i<bitmap.length;i++) // Bitmap
            b[off++]=(byte)bitmap[i];
        for(int i=0;i<768;i++)           // Name table
            b[off++]=(byte)(i&0xff);
        for(int i=0;i<1280;i++)          // Sprite attributes or something
            b[off++]=0;
        for(int i=0;i<colormap.length;i++) // Color table
            b[off++]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Load SC2
    void load_sc2(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<14343)
        {
            message("Can't open file: "+name);
            return;
        }
        message("Loaded: "+name);
        
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        for(int y=0,off=0;y<ysize;y+=8)
        {
            for(int x=0;x<xsize;x+=8)
            {
                off=(((int)b[0x1807+y/8*32+x/8])&0xff)*8+y/64*2048; // Name table
                for(int i=0;i<8;i++)
                {
                    int bits=(int)b[7+off],
                        cols=(int)b[7+6144+768+1280+off];
                        
                    bits&=0xff;
                    cols&=0xff;
                        
                    int fg=(cols>>4), // Extract fg and bg colors
                        bg=(cols&0xf);                        
                    if(fg==0)         // Transparent turns to black
                        fg=1;
                    if(bg==0)
                        bg=1;
                    
                    for(int j=0;j<8;j++)
                    {
                        if((bits&(0x80>>j))!=0)
                            sourcepic.pixels[x+j+(y+i)*xsize]=rgb[fg];
                        else
                            sourcepic.pixels[x+j+(y+i)*xsize]=rgb[bg];
                    }
                    off++;
                }
            }
        }
        
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }

    // Save ColecoVision COL
    void save_col(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/colecovision.col");
        
        populate(); // Create the bitmap/colormap  
        
        // Replace some bytes
        b[12464]=(byte)border;
        for(int i=0;i<bitmap.length;i++)
            b[176+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[6320+i]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }

    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
        
        PrintWriter f=createWriter(name);
        
        f.print("pmap:");
        printarr(f,bitmap,"db","#");
        f.println();
        f.print("cmap:");
        printarr(f,colormap,"db","#");
        f.println();
        f.println("border:\tdb\t#"+hex(border,2));  

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
