/*
    ZX Spectrum specific things
    
    Changelog:
    
    17.5.2016: TAP export
    1.5.2016: Asm export - hopefully some sane format
    30.4.2016: save_scr() and load_scr() implemented
    30.4.2016: Preliminary populate() for file saving
    30.4.2016: messbitmap() and organizebitmap() for translating between linear and Spectrum bitmaps
    29.4.2016: Checker that takes brightness into account
    29.4.2016: Brightness method
    27.4.2016: Removed unused C-64 references
    27.4.2016: First created
*/

class ZXSpectrum extends Machine
{
    // Simple 8-bit RGB with additional brightness
    final int rgb[]={#000000,#0000C0,#C00000,#C000C0,
                     #00C000,#00C0C0,#C0C000,#C0C0C0,
                     #000000,#0000FF,#FF0000,#FF00FF,
                     #00FF00,#00FFFF,#FFFF00,#FFFFFF};
                     
    Button load_scr_b,
           save_scr_b,
           save_tap_b,
           save_asm_b;
    
    ZXSpectrum()
    {
        number=ZX_SPECTRUM;
        name="ZX Spectrum";
        
        xsize=256;
        ysize=192;
        blox=bloy=gridx=gridy=8;

        // Let's just assume they are square pixels        
        renderx=xsize*prefs.zoom;
        rendery=ysize*prefs.zoom;

        colorselheight=prefs.UIWIDTH-2*prefs.BWIDTH;
        
        border=7;
        maxborder=7;
        pen=7;
        
        super.rgb=rgb;
    }
       
    boolean brightness(int colorindex)
    {
        return colorindex>7;
    }
    // Fix all colors to bright/not
    void setbrightness(int pix[],boolean bright)
    {
        for(int i=0;i<pix.length;i++)
        {
            if(bright)
                pix[i]=rgb[bestmatch(pix[i],rgb)|8];
            else
                pix[i]=rgb[bestmatch(pix[i],rgb)&7];
        }
    }
    
    // Mess a linear bitmap to the mess that is the Spectrum bitmap
    void messbitmap()
    {
        int other[]=new int[32*24*8];
        
        for(int y=0,off=0;y<192;y++)
            for(int x=0;x<32;x++,off++)
                other[((y&0xc0)<<5) +((y&0x38)<<2) +((y&7)<<8) +x]=bitmap[off];
        
        bitmap=other;
    }
    // Reorder a Spectrum bitmap to linear
    void organizebitmap()
    {
        int other[]=new int[32*24*8];
        
        for(int y=0,off=0;y<192;y++)
            for(int x=0;x<32;x++,off++)
                other[off]=bitmap[((y&0xc0)<<5) +((y&0x38)<<2) +((y&7)<<8) +x];
            
        bitmap=other;
    }
    
    // Need to account for the brightness
    boolean check(boolean fix)
    {
        boolean errorfound=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        pic.loadPixels();
        for(int y=0;y<ysize;y+=bloy) // Walk through every block
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
                
                if(count.length==2) // Just two, but check brightness
                {
                    if(count[0][0]!=0 && count[1][0]!=0) // With black all is fine
                    {
                        if(brightness(count[0][0])!=brightness(count[1][0]))
                        {
                            errorfound=true;
                            if(fix)
                            {
                                setbrightness(pix,brightness(count[0][0]));
                                setpixels(x,y,blox,bloy,pix);
                            }
                            else
                                setpixels(x,y,blox,bloy,errorcolor);
                        }
                    }
                }
 
                if(count.length>2) // More than two?
                {
                    errorfound=true;
                    if(fix)
                    {
                        // The dominant color decides the brightness unless black
                        if(count[0][0]!=0 && count[0][0]!=0)
                        {
                            setbrightness(pix,brightness(count[0][0]));
                            count=countcolors(pix,rgb); // Recount
                        }

                        if(count.length>1) // Still more than two
                        {
                            // Use the two most common colors for the remaining ones
                            for(int i=0;i<pix.length;i++)
                            {
                                if(rgbdist(pix[i],count[0][1])<rgbdist(pix[i],count[1][1]))
                                    pix[i]=count[0][1];
                                else
                                    pix[i]=count[1][1];
                            }
                            setpixels(x,y,blox,bloy,pix);
                        }
                    }
                    else
                        setpixels(x,y,blox,bloy,errorcolor);
                }
            }
        }
        pic.updatePixels();
        
        if(errorfound)
        {
            message("Color clash(es) found");
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }
    
    // Create color/bitmap in Spectrum format. Needs to be checked before.
    void populate()
    {
        int bx=xsize/blox,
            by=ysize/bloy;
        
        colormap=new int[bx*by];
        bitmap=new int[bx*by*8];
        
        pic.loadPixels();
        for(int y=0,i=0;y<ysize;y+=bloy)
            for(int x=0;x<xsize;x+=blox,i++)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
                
                colormap[i]=count[0][0]&7; // most common color, "ink"
                if(brightness(count[0][0]))
                    colormap[i]|=0x40;
                    
                if(count.length>1) // "paper"
                {
                    colormap[i]+=(count[1][0]&7)<<3;
                    if(brightness(count[1][0]))
                        colormap[i]|=0x40;
                }
                    
                for(int j=0;j<8;j++) // Create the bitmap
                {
                    int bits=0;
                    for(int k=0;k<8;k++)
                        if(pix[j*8+k]==count[0][1])
                            bits+=(0x80>>k);
                    
                    bitmap[x/blox+(y+j)*bx]=bits;
                }
            }

        messbitmap(); // Go Spectrum go
    }
    
    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_scr_b=new Button(x,super.addloadbuttons(x,y),"load scr");
        return load_scr_b.next;
    }
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_scr_b=new Button(x,super.addsavebuttons(x,y),"save scr");
        save_tap_b=new Button(x,save_scr_b.next,"save tap");
        save_asm_b=new Button(x,save_tap_b.next,"save asm");
        return save_asm_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_scr_b.mouseover() || save_tap_b.mouseover() || save_asm_b.mouseover())
        {
            if(pic==null)
                message("Can't save an empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_scr_b.mouseover()) // SCR
                    {
                        String s=fileselector(prefs.path,"SCR files",new String[] {"scr"},SAVE);
                        if(s!=null)
                            save_scr(s);
                    }
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                    if(save_tap_b.mouseover()) // Tap
                    {
                        String s=fileselector(prefs.path,"Tap files",new String[] {"tap"},SAVE);
                        if(s!=null)
                            save_tap(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(load_scr_b.mouseover())
        {
            String s=fileselector(prefs.path,"SCR files",new String[] {"scr"},LOAD);
            if(s!=null)
                load_scr(s);
        }
    }
    
    // Save TAP, a bit idiosyncratic but should work
    void save_tap(String name)
    { 
        if(pic==null)
            return;
          
        byte b[]=loadBytes("templates/spectrum.tap");
        
        populate(); // Create the bitmap/colormap
        
        int headpos,checksum,valu;
        
        checksum=0xc9; // initialize checksum
        for(int i=0x1d24;i<=0x1d3d;i++)
        {
            checksum=checksum^int(b[i]);
        }
        
        //border
        
        b[0x0223]=byte(border);checksum=checksum^int(border);
        headpos=0x0224;
        
        //bitmap
        
        for(int i=0;i<bitmap.length;i++)
        {
            valu=(byte)bitmap[i];
            b[headpos]=byte(valu);checksum=checksum^int(valu);
            headpos++;
        }

        //attribute map
        
        for(int i=0;i<colormap.length;i++)
        {
            valu=(byte)colormap[i];
            b[headpos]=byte(valu);checksum=checksum^int(valu);
            headpos++;
        }
        
        //write in the checksum
        
        headpos=0x1d3e;
        b[headpos]=byte(checksum);
        
        saveBytes(name,b);
        message("Saved: "+name);     
    }
    
    // Save SCR
    void save_scr(String name)
    {
        if(pic==null)
            return;

        byte b[]=new byte[6912];
        
        populate(); // Create the bitmap/colormap
        
        // Replace some bytes
        for(int i=0;i<bitmap.length;i++)
            b[i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[6144+i]=(byte)colormap[i];
        
        saveBytes(name,b);
        message("Saved: "+name);
    }
    // Load SCR
    void load_scr(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<6912)
        {
            message("Can't open file: "+name);
            return;
        }
        message("Loaded: "+name);
        
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        // Fix the bitmap to linear
        bitmap=new int[32*24*8];
        for(int i=0;i<32*24*8;i++)
            bitmap[i]=((int)b[i])&0xff;
        organizebitmap();
        colormap=new int[32*24];
        for(int i=0;i<32*24;i++)
            colormap[i]=((int)b[6144+i])&0x7f; // Flashing will go
        
        for(int y=0,i=0;y<ysize;y+=bloy)
            for(int x=0;x<xsize;x+=blox,i++)
            {
                int ink=colormap[i]&7,
                    paper=(colormap[i]>>3)&7;
                
                if((colormap[i]&0x40)!=0) // Bright
                {
                    paper+=8;
                    ink+=8;
                }
                paper=rgb[paper];
                ink=rgb[ink];
                
                for(int j=0;j<8;j++) // Eight rows with same colors
                {
                    for(int k=0;k<8;k++)
                    {
                        if(((0x80>>k)&bitmap[x/blox+(j+y)*32])!=0)
                            sourcepic.pixels[x+k+(y+j)*xsize]=ink;
                        else
                            sourcepic.pixels[x+k+(y+j)*xsize]=paper;
                    }
                }
            }
           
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }

    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
       
        PrintWriter f=createWriter(name);
        
        f.print("bmap");
        printarr(f,bitmap,"defb","$");
        f.println();
        f.print("cmap");
        printarr(f,colormap,"defb","$");
        f.println();
        f.println("border\tdefb\t$"+hex(border,2));

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
