/*
    Base class for all the machines
    
    Changelog:
    
    18.5.2016: Panasonic JR-200 joins the crew
    1.5.2016: Added maxborder to limit possible border colors
    29.4.2016: Improved check() that uses the two most common colors, moved bestmatch() to tools
    29.4.2016: Auto forcing is now possible too
    28.4.2016: Convenience functions bestmatch(array) and get/setpixels(). Simplified check() loop.
    28.4.2016: Accept wrong size images and scale them
    28.4.2016: Added gridx/gridy, since the grid isn't necessarily the same as the block size
    28.4.2016: Take blox/bloy into account with the grid, renumbered the platforms
    28.4.2016: check() returns true if clashes not found, false otherwise
    27.4.2016: First attempt at automatic polling of file changes
    27.4.2016: New method logic() for doing something every frame
    27.4.2016: Zooming added
    27.4.2016: Do not smooth zoomed images to keep the grid aligned with the blocks
    27.4.2016: Simple x*y block color checker/fixer done, see check()
    27.4.2016: More dummy buttons (force and check), added pic to keep the original intact
    27.4.2016: Grid drawing and color selector clicking work
    27.4.2016: Loading and saving PNGs works
    27.4.2016: Button add methods, first file selector, colorcorrect()
    26.4.2016: The first version
*/

// Machine number defines
final int C64_HIRES=0,
          C64_MULTICOLOR=1,
          CPC_MODE0=2,
          CPC_MODE0OVER=3,
          CPC_MODE1=4,
          CPC_MODE1OVER=5,
          MSX_SCREEN2=6,
          MSX_SCREEN3=7,
          ORIC=8,
          PANASONIC_JR200=9,
          PLUS4_HIRES=10,
          PLUS4_MULTICOLOR=11,
          SHARP_MZ700=12,
          SHARP_MZ800=13,
          ZX_SPECTRUM=14;

class Machine
{
    int number,
    
        xsize,ysize,     // Actual pixels
        renderx,rendery, // Shown on screen
        colorselheight,
        blox,bloy,       // "Character" block for clash checking
        gridx,gridy,     // Same for the grid - might not be the same (for example the MSX)
        border,
        maxborder,
        pen,
        
        autocheck=0,
        
        gridcolor=#603030,  // Just a guess, machines can set it to something else
        errorcolor=#706030; // Offending block color, machine-specific as well (visible and not one of the set)
        
    long oldstamp=-1;
        
    String name;
    
    int rgb[],
        colormap[],        // These two for saving the file
        bitmap[];
    
    PImage sourcepic=null, // Loaded piccy
           pic=null;       // Converted etc. piccy
    
    boolean grid=false,
            zoom=false;
    
    Button load_image_b,
           save_image_b,
           check_b,
           revert_b,
           force_b;
           
    Checkbutton grid_b,
                zoom_b,
                autocheck_b,
                autoforce_b;
    
    Machine()
    {
    }
    
    // Things to do every frame
    void logic()
    {
        if(autocheck>0 && frameCount%60==0)
        {
            // Check if file modification time has changed
            long stamppi=timestamp(prefs.filename);
            if(stamppi!=oldstamp)
            {
                // Give it time to settle... Yes, it's a kludge :/
                do
                {
                    stamppi=timestamp(prefs.filename);
                    delay(500);
                } while(stamppi!=timestamp(prefs.filename));
                
                load_image(prefs.filename);
                check(autocheck==2);
                oldstamp=stamppi;
                
                repaint=true;
            }
        }
    }
    
    // Update the canvas
    void drawcanvas()
    {
        if(pic==null)
        {
            fill(rgb[machine.pen]);
            rect(canvas_startx,canvas_starty, renderx,rendery);
        }
        else
        {
            noSmooth();
            if(incanvas() && zoom) // Zoom while on canvas
            {
                int sx=(mouseX-canvas_startx)*xsize/renderx/2,
                    sy=(mouseY-canvas_starty)*ysize/rendery/2;
                copy(pic,sx,sy,pic.width/2,pic.height/2, canvas_startx,canvas_starty, renderx,rendery);
            }
            else // Just copy the image
            {
                image(pic,canvas_startx,canvas_starty, renderx,rendery);
            }
            smooth();
        }
            
        stroke(gridcolor);
        if(grid) // More complicated grid when zooming
        {
            if(incanvas() && zoom)
            {
                // Ended up quite convoluted, but works almost enough: move the grid in the zoomed mode
                int sx=-((mouseX-canvas_startx)*xsize/renderx/2)*2*(renderx+1)/xsize,
                    sy=-((mouseY-canvas_starty)*ysize/rendery/2)*2*(rendery+1)/ysize;
                
                float mul=(float)rendery*2/(float)ysize;
                for(int y=0;y<=ysize;y+=gridy)
                {
                    int yy=sy+canvas_starty+(int)(mul*(float)y);
                    if(yy>=canvas_starty && yy<canvas_endy)
                        line(canvas_startx,yy,canvas_endx,yy);
                }
                         
                mul=(float)renderx*2/(float)xsize;
                for(int x=0;x<=xsize;x+=gridx)
                {
                    int xx=sx+canvas_startx+(int)(mul*(float)x);
                    if(xx>=canvas_startx && xx<canvas_endx)
                        line(xx,canvas_starty,xx,canvas_endy);
                }
                
                noFill();
                rect(canvas_startx,canvas_starty, canvas_endx-canvas_startx,canvas_endy-canvas_starty);
            }
            else // Plain old grid
            {
                float mul=(float)rendery/(float)ysize;
                for(int y=0;y<=ysize;y+=gridy)
                    line(canvas_startx,canvas_starty+(int)(mul*(float)y),
                           canvas_endx,canvas_starty+(int)(mul*(float)y));
                         
                mul=(float)renderx/(float)xsize;
                for(int x=0;x<=xsize;x+=gridx)
                    line(canvas_startx+(int)(mul*(float)x),canvas_starty,
                         canvas_startx+(int)(mul*(float)x),canvas_endy);
            }
        }
        noStroke();
    }
    
    // Draw the color selector
    void drawcolorsel(int px,int py)
    {
        int wide=(prefs.UIWIDTH-2*prefs.BWIDTH)/4,
            divisor=4,
            selectedx=0,
            selectedy=0,
            selected=0;
            
        if(rgb.length>16)
        {
            wide/=2;
            divisor=8;
        }
        
        for(int y=0,i=0;y<rgb.length/divisor;y++)
        {
            for(int x=0;x<divisor;x++,i++)
            {
                if(rgb[i]!=ILLEGAL)
                {
                    fill(rgb[i]);
                    rect(px+x*wide,py+y*wide, wide,wide);
                    
                    if(mousePressed && !shift && mouseButton==LEFT) // A bit of a hack, yes
                    {
                        if(inside(px+x*wide,py+y*wide, px+x*wide+wide,py+y*wide+wide))
                            if(i<=maxborder)
                                border=i;
                    }
                }
                else // Not selectable
                {
                    stroke(200,0,0,230);
                    line(px+x*wide,py+y*wide, px+x*wide+wide,py+y*wide+wide);
                    line(px+x*wide+wide,py+y*wide, px+x*wide,py+y*wide+wide);
                    noStroke();
                }
                
                if(i==border)
                {
                    selected=i;
                    selectedx=px+x*wide;
                    selectedy=py+y*wide;
                }
            }
        }
        
        noFill();
        strokeWeight(2);
        stroke(rgb[selected]^0x808080);
        rect(selectedx,selectedy, wide,wide);
        strokeWeight(1);
        noStroke();
        
        fill(prefs.UITEXT);
        text("border: "+str(border),px,py+wide*rgb.length/divisor+prefs.UIROW-7);
    }
    
    // Utility functions:
    // Get a block of pixels as an array. Better do loadPixels() before this.
    int[] getpixels(int x,int y,int w,int h)
    {
        if(x<0 || y<0 || x+w>xsize || y+h>ysize)
            return null;
        
        int arr[]=new int[w*h];
        
        for(int i=0;i<h;i++)
            for(int j=0;j<w;j++)
                arr[i*w+j]=pic.pixels[x+j+pic.width*(y+i)];
        
        return arr;
    }
    // Set a block of pixels from an array. Remember loadPixels() first
    boolean setpixels(int x,int y,int w,int h,int pix[])
    {
        if(x<0 || y<0 || x+w>xsize || y+h>ysize || pix.length<w*h)
            return false;
        
        for(int i=0;i<h;i++)
            for(int j=0;j<w;j++)
                pic.pixels[x+j+pic.width*(y+i)]=pix[i*w+j];
                
        return true;
    }
    // Set a block of pixels the same color. Remember loadPixels() first
    boolean setpixels(int x,int y,int w,int h,int kolor)
    {
        if(x<0 || y<0 || x+w>xsize || y+h>ysize)
            return false;
        
        for(int i=0;i<h;i++)
            for(int j=0;j<w;j++)
                pic.pixels[x+j+pic.width*(y+i)]=kolor;
                
        return true;
    }
    
    // Map loaded pic to machine colors in pic - no other limit checking yet
    void colorcorrect()
    {
        if(pic==null || sourcepic==null)
            return;
        
        pic.loadPixels();
        sourcepic.loadPixels();
        for(int i=0;i<sourcepic.width*sourcepic.height;i++)
            pic.pixels[i]=rgb[bestmatch(sourcepic.pixels[i],rgb)];
        pic.updatePixels();
    }
    
    // Check two-color limits and fix if needed. Make a different logic if the machine requires so.
    boolean check(boolean fix)
    {
        boolean errorfound=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        pic.loadPixels();
        for(int y=0;y<ysize;y+=bloy) // Walk through every block
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
 
                if(count.length>2) // More than two?
                {
                    errorfound=true;
                    if(fix)
                    {
                        // Use the two most common colors for the remaining ones
                        for(int i=0;i<pix.length;i++)
                        {
                            if(rgbdist(pix[i],count[0][1])<rgbdist(pix[i],count[1][1]))
                                pix[i]=count[0][1];
                            else
                                pix[i]=count[1][1];
                        }
                        setpixels(x,y,blox,bloy,pix);
                    }
                    else
                        setpixels(x,y,blox,bloy,errorcolor);
                }
            }
        }
        pic.updatePixels();
        
        if(errorfound)
        {
            message("Color clash(es) found");
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }
    
    // Check image dimensions
    boolean correctsize(int x,int y)
    {
        if(x==xsize && y==ysize)
            return true;
        else
            return false;
    }
    
    // Add all the load buttons and return the next row
    int addloadbuttons(int x,int y)
    {
        load_image_b=new Button(x,y,"load png/gif");
        return load_image_b.next;
    }
    // Add all the save buttons and return the next row
    int addsavebuttons(int x,int y)
    {
        save_image_b=new Button(x,y,"save png");
        return save_image_b.next;
    }
    // Add all extra buttons and return the next row
    int addextrabuttons(int x,int y)
    {
        grid_b=new Checkbutton(x,y,"grid");
        zoom_b=new Checkbutton(x,grid_b.next,"zoom");
        check_b=new Button(x,zoom_b.next+prefs.UIGAP,"check");
        autocheck_b=new Checkbutton(x,check_b.next,"auto check");
        force_b=new Button(x,autocheck_b.next,"force");
        autoforce_b=new Checkbutton(x,force_b.next,"auto force");
        revert_b=new Button(x,autoforce_b.next,"revert");
        return revert_b.next;
    }

    // Buttons pressed?    
    void handleclicks()
    {
        if(load_image_b.mouseover())
        {
            String s=fileselector(prefs.path,"PNG/GIF images",new String[] {"gif","png"},LOAD);
            if(s!=null)
                load_image(s);
        }
        if(save_image_b.mouseover())
        {
            if(pic==null)
            {
                message("Empty file can't be saved");
            }
            else
            {
                String s=fileselector(prefs.path,"PNG images",new String[] {"png"},SAVE);
                if(s!=null)
                    save_image(s);
            }
        }
        
        if(check_b.mouseover())
        {
            colorcorrect();
            check(false);
        }
        if(revert_b.mouseover())
        {
            colorcorrect();
        }
        if(force_b.mouseover())
        {
            colorcorrect();
            check(true);
        }
        
        if(zoom_b.mouseover())
        {
            zoom=!zoom;
            zoom_b.checked=zoom;
        }
        if(grid_b.mouseover())
        {
            grid=!grid;
            grid_b.checked=grid;
        }
        if(autocheck_b.mouseover() || autoforce_b.mouseover())
        {
            if(pic!=null)
            {
                if(getext(prefs.filename).equalsIgnoreCase(".png") || getext(prefs.filename).equalsIgnoreCase(".gif"))
                {
                    if(autocheck_b.mouseover())
                    {
                        if(autocheck==1)
                            autocheck=0;
                        else
                            autocheck=1;
                    }
                    else // Autoforce
                    {
                        if(autocheck==2)
                            autocheck=0;
                        else
                            autocheck=2;
                    }
                        
                    autocheck_b.checked=autocheck==1;
                    autoforce_b.checked=autocheck==2;
                 
                    if(autocheck>0)
                    {
                        colorcorrect();
                        check(autocheck==2);
                    }
                }
                else
                    message("Only PNG/GIF images can be autochecked");
            }
            else
                message("Can't check an empty image");
        }
    }
    
    // File operations:
    // Load PNG/GIF
    void load_image(String name)
    {
        PImage tmp=loadImage(name);
        if(tmp==null)
            message("Can't open file: "+name);
        else
        {
            if(!correctsize(tmp.width,tmp.height))
                message("Warning: non-native size "+str(tmp.width)+"x"+str(tmp.height));
            else
                message("Loaded: "+name);
                
            sourcepic=tmp;
            prefs.filename=name;
            
            if(pic==null)
                pic=new PImage(xsize,ysize,RGB);
             
            // Scale it then
            if(tmp.width!=xsize || tmp.height!=ysize)
            {
                // This mess is needed because PImage.copy() always does antialiasing, copy() not
                sourcepic=new PImage(xsize,ysize);
                noSmooth();
                copy(tmp, 0,0,tmp.width,tmp.height, 0,0,xsize,ysize);    
                smooth();
                loadPixels();
                sourcepic.loadPixels();
                for(int y=0;y<ysize;y++)
                {
                    for(int x=0;x<xsize;x++)
                        sourcepic.pixels[y*xsize+x]=pixels[y*width+x];
                }
                sourcepic.updatePixels();
            }
            
            colorcorrect();
        }
        
        oldstamp=-1;
    }
    // Save PNG
    void save_image(String name)
    {
        if(pic==null)
            return;
        
        if(getext(name)==null)
            name=name+".png";
        if(!getext(name).equalsIgnoreCase(".png"))
            name=name+".png";
        
        if(pic.save(name))
        {
            message("Saved: "+name);
            prefs.filename=name;
        }
        else
            message("Can't save file: "+name);
    }
}
